// File: DAGAd.h
// Author: Francesco Giacomini <Francesco.Giacomini@cnaf.infn.it>
// Copyright (c) 2001 EU DataGrid.
// For license conditions see http://www.eu-datagrid.org/license.html

// $Id: DAGAd.h,v 1.9 2005/07/04 14:57:17 amarasch Exp $

#ifndef GLITE_WMS_COMMON_REQUESTAD_DAGAD_H
#define GLITE_WMS_COMMON_REQUESTAD_DAGAD_H

#include <string>
#include <utility>
#include <map>
#include <set>
#include <iosfwd>
#include <classad_distribution.h>
#include <boost/shared_ptr.hpp>
#include <boost/utility.hpp>
#include <exception>
#include "glite/wms/common/utilities/classad_utils.h"

//graph libraries
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/depth_first_search.hpp>
#include <boost/graph/visitors.hpp>

namespace glite {
namespace wms {
namespace jdl {

namespace utilities = glite::wms::common::utilities;

struct InvalidDAG: public std::exception
{
  char const* what() const throw()
  {
    return "Invalid DAG";
  }
};

struct InvalidNodeInfo: public std::exception
{
  char const* what() const throw()
  {
    return "Invalid Node Info";
  }
};

class DAGNodeInfo
{
  class Implementation;
  boost::shared_ptr<Implementation> m_impl;

public:
  DAGNodeInfo();
  explicit DAGNodeInfo(classad::ClassAd const& node_info_ad);
  DAGNodeInfo(std::string const& node_description,
              std::string const& node_type);
  DAGNodeInfo(classad::ClassAd const& node_description,
              std::string const& node_type);

public:
  std::string             description_file() const;
  bool                    replace_description_file(std::string const& file);
  classad::ClassAd const* description_ad() const;
  bool                    replace_description_ad(classad::ClassAd* ad);
  bool                    description_file_for_ad(std::string const& file);
  bool                    description_ad_for_file(classad::ClassAd* ad);

  std::string             type() const;

  std::pair<std::string,std::string> pre() const;
  bool                    pre(std::string const& file, std::string const& args);
  std::pair<std::string,std::string> post() const;
  bool                    post(std::string const& file, std::string const& args);

  int                     retry_count() const;
  bool                    retry_count(int n);

  classad::ClassAd        as_classad() const;
};

class DAGAdNodeIterator
  : public std::iterator<
  std::input_iterator_tag
  ,std::pair<std::string, DAGNodeInfo>
  ,ptrdiff_t
  ,std::pair<std::string, DAGNodeInfo> const*
  ,std::pair<std::string, DAGNodeInfo> const&
>
{
  friend bool operator==(DAGAdNodeIterator const& lhs, DAGAdNodeIterator const& rhs);

  classad::ClassAd const* m_nodes;
  classad::ClassAd::const_iterator m_node;
  value_type m_value;

  bool is_good() const
  {
    return m_node != m_nodes->end();
  }

  void set_value();

public:

  DAGAdNodeIterator();
  ~DAGAdNodeIterator();
  DAGAdNodeIterator(classad::ClassAd const* nodes, classad::ClassAd::const_iterator it);
  DAGAdNodeIterator(DAGAdNodeIterator const& other);
  DAGAdNodeIterator& operator=(DAGAdNodeIterator const& other);
  DAGAdNodeIterator& operator++();
  DAGAdNodeIterator operator++(int);
  reference operator*() const;

  pointer operator->() const
  {
    return &(operator*());
  }

};

inline bool operator==(DAGAdNodeIterator const& lhs, DAGAdNodeIterator const& rhs)
{
  return lhs.m_nodes == rhs.m_nodes && lhs.m_node == rhs.m_node;
}

inline bool operator!=(DAGAdNodeIterator const& lhs, DAGAdNodeIterator const& rhs)
{
  return !(lhs == rhs);
}

struct DAGAdDependencyIterator
  : public std::iterator<
  std::input_iterator_tag
  ,std::pair<DAGAdNodeIterator, DAGAdNodeIterator>
  ,ptrdiff_t
  ,std::pair<DAGAdNodeIterator, DAGAdNodeIterator> const*
  ,std::pair<DAGAdNodeIterator, DAGAdNodeIterator> const&
>
{
  friend bool operator==(DAGAdDependencyIterator const& lhs,
                         DAGAdDependencyIterator const& rhs);

  typedef DAGAdNodeIterator node_iterator;
  typedef classad::ExprList::iterator Iterator;

  classad::ClassAd const* m_dag;
  Iterator m_dep;
  Iterator m_dep_end;
  Iterator m_parent;
  Iterator m_child;
  value_type m_value;

  static classad::ExprList* parents(classad::ExprTree* dep)
  {
    assert(utilities::is_expression_list(dep));
    classad::ExprTree* et = *static_cast<classad::ExprList*>(dep)->begin();
    assert(utilities::is_expression_list(et));
    return static_cast<classad::ExprList*>(et);
  }
  static classad::ExprList* children(classad::ExprTree* dep)
  {
    assert(utilities::is_expression_list(dep));
    classad::ExprTree* et = *++static_cast<classad::ExprList*>(dep)->begin();
    assert(utilities::is_expression_list(et));
    return static_cast<classad::ExprList*>(et);
  }
  bool is_good() const { return m_dep != m_dep_end; }
  void set_value();

  DAGAdDependencyIterator(classad::ClassAd const* dag = 0, bool begin = false);

  // copy ctor, operator=() and dtor are ok

  DAGAdDependencyIterator& operator++();
  DAGAdDependencyIterator operator++(int);

  reference operator*() const
  {
    return m_value;
  }

  pointer operator->() const
  {
    return &(operator*());
  }
};

inline bool operator==(DAGAdDependencyIterator const& lhs,
                       DAGAdDependencyIterator const& rhs)
{
  return (!lhs.is_good() && !rhs.is_good())
    || (lhs.is_good() && rhs.is_good()
        && lhs.m_dep == rhs.m_dep
        && lhs.m_parent == rhs.m_parent
        && lhs.m_child == rhs.m_child
        );
}

inline bool operator!=(DAGAdDependencyIterator const& lhs,
                DAGAdDependencyIterator const& rhs)
{
  return !(lhs == rhs);
}

class DAGAd: boost::noncopyable
{
public:

  typedef DAGAdNodeIterator node_iterator;
  typedef DAGAdNodeIterator::value_type node_value_type;
  typedef DAGAdDependencyIterator dependency_iterator;
  typedef DAGAdDependencyIterator::value_type dependency_value_type;
  typedef boost::adjacency_list<boost::vecS, boost::vecS, boost::directedS> Graph_t;
  typedef boost::graph_traits<Graph_t>::vertex_descriptor Vertex;
  typedef boost::property_map<Graph_t, boost::vertex_color_t>::type Color;
  typedef std::map<std::string, Vertex> Mapping;

private:
  boost::shared_ptr<classad::ClassAd> m_ad;
  //graph variables
  Graph_t m_graph;
  Mapping map_nodes;


public:
  DAGAd();
  explicit DAGAd(classad::ClassAd const& ad);

  classad::ClassAd const& ad() const;

  // set the maximum number of running nodes
  // if new_value is -1 the attribute becomes unspecified
  // return the previous value (-1 means unspecified)
  int max_running_nodes(int new_value);

  // get the maximum number of running nodes
  // 0 means unspecified
  int max_running_nodes() const;

  // set the default node type
  // if new_value is "" the attribute becomes unspecified
  // return the previous value ("" means unspecified)
  std::string default_node_type(std::string const& new_value);

  // get the default node type
  // "" means unspecified
  std::string default_node_type() const;

  // set the default retry count for a node
  // if new_value is -1 the attribute becomes unspecified
  // return the previous value (-1 means unspecified)
  int default_node_retry_count(int new_value);

  // get the default retry count for a node
  // -1 means unspecified
  int default_node_retry_count() const;

  bool add_node(std::string const& name, DAGNodeInfo const& info);
  bool replace_node(std::string const& name, DAGNodeInfo const& info);
  bool remove_node(std::string const& name);
  std::size_t num_nodes() const;
  node_iterator find(std::string const& name) const;

  bool add_dependency(std::string const& first, std::string const& second);
  bool remove_dependency(std::string const& first, std::string const& second);
  std::size_t num_dependencies() const;

  bool set_generic(std::string const& attribute, classad::ExprTree* value);
  classad::ExprTree const* get_generic(std::string const& attribute) const;
  bool remove_generic(std::string const& attribute);

  std::pair<dependency_iterator, dependency_iterator> dependencies() const;
  std::pair<node_iterator, node_iterator> nodes() const;

  struct Attributes
  {
    static std::string const TYPE;
    static std::string const NODES;
    static std::string const DEPENDENCIES;
    static std::string const MAX_RUNNING_NODES;
    static std::string const NODE_RETRY_COUNT;
    static std::string const NODE_TYPE;
    static std::string const DESCRIPTION_FILE;
    static std::string const DESCRIPTION_AD;
    static std::string const PRE;
    static std::string const PRE_ARGUMENTS;
    static std::string const POST;
    static std::string const POST_ARGUMENTS;
  };
};

} // jdl namespace
} // wms namespace
} // glite namespace

#endif
